import importlib
from typing import Any


def dynamic_import(name: str) -> Any:
    components = name.split(".")
    variable = components[-1]
    mod = importlib.import_module(".".join(components[:-1]))
    return getattr(mod, variable)



import inspect

import click

from .utils import dynamic_import
from ..environment import EnvironmentConfig


@click.group()
def main() -> None:
    pass


@main.group()
def env() -> None:
    pass


@env.command("show")
@click.option(
    "--config-module-path",
    required=True,
    help="python module path to root environment config",
)
def show_env(config_module_path: str) -> None:
    ConfigClass = dynamic_import(config_module_path)
    if not inspect.isclass(ConfigClass) or not issubclass(
        ConfigClass, EnvironmentConfig
    ):
        raise click.UsageError(
            f'"{config_module_path}" is not valid '
            "subclass of EnvironmentConfig"
        )
    variables = ConfigClass.get_all_variables()
    for v in variables:
        click.echo(f"\"{v.key}\" {'optional' if v.optional else 'required'}")



from abc import ABC, abstractmethod

from .lazy_value import LazyValue


class IEnvironmentManager(ABC):
    @abstractmethod
    def get(self, key: str) -> LazyValue:
        raise NotImplementedError



from __future__ import annotations

import inspect
from typing import Any, Iterable, Type

from .env_conf_value import EnvConfValue
from ..imanager import IEnvironmentManager


class EnvironmentConfig:
    def __init__(self, environment: IEnvironmentManager) -> None:
        self.load_values(environment)
        self.load_configs(environment)

    def load_values(self, environment: IEnvironmentManager) -> None:
        for value in self.get_env_values():
            value.load_value(environment)

    def load_configs(self, environment: IEnvironmentManager) -> None:
        for key, ConfigClass in self.get_configs():
            setattr(self, key, ConfigClass(environment))

    @classmethod
    def get_env_values(cls) -> Iterable[EnvConfValue[Any]]:
        return filter(
            lambda value: isinstance(value, EnvConfValue),
            cls.__dict__.values(),
        )

    @classmethod
    def get_configs(cls) -> Iterable[tuple[str, Type[EnvironmentConfig]]]:
        return filter(
            lambda item: inspect.isclass(item[1])
            and issubclass(item[1], EnvironmentConfig),
            inspect.get_annotations(cls, eval_str=True).items(),
        )

    @classmethod
    def get_all_variables(cls) -> list[EnvConfValue[Any]]:
        variables = list(cls.get_env_values())
        for _, ConfigClass in cls.get_configs():
            variables += ConfigClass.get_all_variables()
        return variables



from __future__ import annotations

from typing import (
    Any,
    Callable,
    Generic,
    Literal,
    TypeVar,
    overload,
)

from ..imanager import IEnvironmentManager


T = TypeVar("T")


class EnvConfValue(Generic[T]):
    key: str
    optional: bool
    default: str | None
    _loaded: bool
    _value: T | None
    validator: Callable[[str], None] | None
    converter: Callable[[str], Any] | None

    @overload
    def __init__(
        self: EnvConfValue[str],
        key: str,
        default: str | None = None,
        optional: Literal[False] = False,
        validator: Callable[[str], None] | None = None,
        converter: Callable[[str], str] | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self: EnvConfValue[str | None],
        key: str,
        default: None = None,
        optional: Literal[True] = True,
        validator: Callable[[str], None] | None = None,
        converter: Callable[[str], str | None] | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self: EnvConfValue[T],
        key: str,
        default: str | None = None,
        optional: Literal[False] = False,
        validator: Callable[[str], None] | None = None,
        converter: Callable[[str], T] | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self: EnvConfValue[T | None],
        key: str,
        default: None = None,
        optional: Literal[True] = True,
        validator: Callable[[str], None] | None = None,
        converter: Callable[[str], T] | None = None,
    ) -> None:
        pass

    def __init__(
        self: EnvConfValue[T | str | None],
        key: str,
        default: str | None = None,
        optional: bool = False,
        validator: Callable[[str], None] | None = None,
        converter: Callable[[str], T | str | None] | None = None,
    ) -> None:
        self.key = key
        self.default = default
        self.optional = optional
        self.validator = validator
        self.converter = converter
        self._loaded = False
        self._value = None

    def load_value(self, environment: IEnvironmentManager) -> None:
        loader = environment.get(self.key)
        if self.optional:
            loader.optional()
        if self.validator:
            loader.validator(self.validator)
        if self.converter:
            loader.converter(self.converter)
        if self.default:
            loader.default(self.default)
        self._value = loader.resolve()
        self._loaded = True

    @property
    def value(self) -> T:
        if not self._loaded:
            raise ValueError(f'EnvConfValue "{self.key}" not loaded value yet')
        return self._value  # type: ignore [return-value]
